//********************************************************************
//*** ResourcePool/BatchCommand.hh
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: BatchCommand.hh,v 1.20 2009-05-11 14:06:05 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#ifndef RESOURCEPOOL_BATCHCOMMAND_HH
#define RESOURCEPOOL_BATCHCOMMAND_HH

#include <vector>

#include "mm/gc_ptr.hh"
#include "util/Clone.hh"

namespace fatalmind {

/**
 * The class is a collection of Commands to be executed in sequential order.
 * Execution of commands stopps in case an exception is thrown.
 * BatchCommand's can also be nested.
 */
template<class CommandType>
class BatchCommand: public CommandType
{
	protected:
		typedef gc_ptr<CommandType> member_t;
		typedef std::vector<member_t> list_t;
	public:
		typedef CommandType innercommand_t;
		typedef typename list_t::const_iterator const_iterator;
		static const int NOT_STARTED;
		static const int NO_FURTHER_ERROR;

		/**
		 * Creates a new (empty) Batch of commands.
		 * The cont flag specifies if the continuation
		 * mode will be active for this batch.
		 *
		 * Per default the batch can be executed several time
		 * and will always start from beginning (even in case 
		 * a previous execution stoped because of an error).
		 *
		 * In continuation mode repetative executons will
		 * continue the batch's execution after the position
		 * where the last execution stoped.
		 * This is useful to continue execution after an error.
		 * The command which caused the error will not be
		 * re-executed.
		 * If all batched commands were executed subsequent
		 * executions of the batch will do nothing.
		 *
		 * When using the continuation mode it is required
		 * to execute the batch until it completed without error.
	 	 * Otherwise it is undefined which statements were acutally
		 * executed and which not.
		 *
		 * see BatchCommand::getExecPosition()
		 * see BatchCommand::getError()
		 */
		BatchCommand(bool cont = false);

		BatchCommand(const BatchCommand&);
		virtual ~BatchCommand() {
		}

		/**
		 * Adds a command to the batch.
		 *
		 * The command will be added to the end of the batch 
		 * like push_back for vectros.
		 *
		 * Please note that adding commands to an Batch is not
		 * allowed after the batch was executed(), not even if 
		 * the execution aborted due to an exception.
	 	 */
		virtual void add(const CommandType&);
		virtual void execute(typename CommandType::pool_type::resource_t&);

		/**
		 * Returns the number of the command which failed during execution.
		 * 
		 * Can be used together with BatchCommand::begin() to obtain the command which failed:
		 * \code
		 * int error = batchcommand.getError();
		 * if (error >= 0) {
		 *    Command& failed = *(batchcommand.begin() + error);
		 * }
		 * \endcode
		 * 
		 * In continuation mode (see BatchCommand::BatchCommand()) getError()
		 * will return all errors, one for every call. To fetch all errors use
		 * a structure like this:
		 * 
		 * \code
		 * while (b.getExecPosition() != b.end()) {
		 *   try {
		 *     p.execute(b);
		 *   } catch (...) {
		 *     // each unsuccessful execution will deliver one error
		 *     p::Command::Base& failed = *(b.begin() + b.getError());
		 *   }
		 * }
		 * \endcode
		 *
		 * The getError() method returns only the errors which already happend.
		 * After unsuccessful execution of a batch, it is guaranteed that getError()
		 * one error, after the next unsuccessful execution the next will be available.
		 *
		 * You could also execute the batch first (until no error) and then fetch all
		 * failed commands.
		 *
		 * \retval >=0 Index of a failed command
		 * \retval <0 non-error conditions. Listed below
		 * \retval BatchCommand::NOT_STARTED in case the execution was not yet started
		 * \retval BatchCommand::NO_FURTHER_ERROR in case no further errors are available (yet)
		 */
		int getError();
		/**
		 * returns the current execution position.
		 * required to determine if the batch has been executed completley.
		 * if equals end() the batch was fully executed.
		 */
		const_iterator getExecPosition() const;

		/**
		 * A Predicate if the continuation mode is enabled in this command.
		 * \retval true If the continuation mode is enabled
		 * \retval false otherwise
		 */
		bool doesContinue() const {
			return cont;
		}
		
		/**
		 * Returns an iterator beyond the last command in the batch.
		 *
		 */
		const_iterator end() const;
		/**
		 * Returns an interator to the first command in the batch.
		 *
		 * will equals to end() if the batch is empty.
		 */
		const_iterator begin() const;
		/**
		 * \retval true if the batch is empty.
		 * \retval else if there is at lease one command in the batch.
		 */
		bool empty() const;

		virtual void outputoperator(std::ostream&) const;
	protected:
		virtual Clone* DoClone() const {
			return new BatchCommand(*this);
		}

		member_t getLast() const {
			return list.back();
		}

		bool wasStarted() const {
			return started;
		}

		
	private:
		list_t list;
		const bool cont;
		bool started;
	protected:
		const_iterator exec_pos;
		typedef std::vector<int> errors_t;
		errors_t errors;
		unsigned int error_pos;
}; // class Batch

} // namespace fatalmind
#endif
