//********************************************************************
//*** ResourcePool/ResourcePool.cc
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: ResourcePool.cc,v 1.39 2009-03-18 16:36:57 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#ifndef RESOURCEPOOL_RESOURCEPOOL_HH
#include "ResourcePool/ResourcePool.hh"
#endif

#ifndef RESOURCEPOOL_CREATEEXCEPTION_HH
#include "CreateException.hh"
#endif

#include "ResourcePool/NoFailoverException.hh"


#include <sstream>

namespace fatalmind {

template<class Type>
ResourcePool<Type>::ResourcePool(
	  const typename Type::factory_t& a_factory
	, const ResourcePoolOptions& a_options
//	, ResourcePoolEventObserver& log
): ResourcePoolInterface<Type>()
	, _factory(a_factory)
	, _options(a_options)
	, _stopped(false)
	, observer(_factory.getEventObserver())
//log.clone<ResourcePoolEventObserver>())
{
	try {
		for (int i = _options.getPreCreate(); i > 0; --i) {
			_inc_pool();
		}
	} catch (const CreateException& e) {
	}
}

template<class Type>
ResourcePool<Type>::~ResourcePool()
{
	sync_t lk(_condvar);
	_stopped = true;
	_downsize();
	if (TM::isMultiThreaded) {
		while (!_usedPool.empty()) {
			_condvar.wait();
			_downsize();
		}
	}
}

template<class Type>
void ResourcePool<Type>::_downsize()
{
	event(ResourcePoolEventObserver::Event.DOWNSIZE_POOL);
	typename freePool_t::iterator i = _freePool.begin();
	typename freePool_t::iterator end = _freePool.end();
	for (; i != end; ++i) {
		(*i)->close();
	}
	_freePool.clear();
}

template<class Type>
void ResourcePool<Type>::downsize()
{
	sync_t lk(_condvar);
	_downsize();
}

template<class Type>
ResourcePoolStats ResourcePool<Type>::getStats() const
{
	sync_t lk(_condvar);
	// should be safe, i guess, ISO C++ 12.2
	return ResourcePoolStats(_stats, _usedPool.size(), _freePool.size()); 
}

template<class Type>
typename Type::resource_t& ResourcePool<Type>::get()
{
	typename Type::resource_t* rec = 0;
	int maxtry = _options.getMaxTry() - 1;

	do {
		try {
			rec = moveToUsed();
			if (! rec->precheck()) { // TODO, what happens in case of exception in prechec/postcheck?
				sync_t lk(_condvar);
				rec->fail_close();
				_usedPool.erase(plainresource_ptr(*rec));
				rec = 0;
				throw CreateException("precheck failed for a resource from the pool");
			}
		} catch (const CreateException& e) {
			if (maxtry > 0) {
				sleepit(maxtry-1);
			} else {
				event(ResourcePoolEventObserver::Event.GET_RESOURCE_FAIL);
				throw;
			}
		}
	} while ((rec == 0) && (maxtry-- > 0));

	if (rec != 0) {
		return *rec;
	} else {
		// acutally impossible?
		abort();
	}
}

// returns direct reference
template<class Type>
typename Type::resource_t* ResourcePool<Type>::moveToUsed() {
	sync_t lk(_condvar);
	_stats.getCall_();

	if (_stopped) {
		throw CreateException("This ResourcePool is beeing stopped");
	}
	// TODO: make noWait() and Wait(time) in options
	while (_freePool.empty()) {
		try {
			_inc_pool();
		} catch (const CreateException& e) {
			_stats.waitEvent_();
			_waitForFree(e, typename TM::Type());
		}
	}
	
	resource_ptr rec(_freePool.front());
	std::pair<typename usedPool_t::const_iterator, bool> rv = _usedPool.insert(std::make_pair(plainresource_ptr(*rec), rec));
	_freePool.pop_front();
	return &*rec;
}

template<class Type>
void ResourcePool<Type>::_waitForFree(
	  const CreateException& e
	, const SingleThreadedModel<>::Type t
) {
	throw e;	
}

template<class Type>
void ResourcePool<Type>::_waitForFree(
	  const CreateException& e
	, const MultiThreadedModel<>::Type t
) {
	if (! _condvar.wait(_options.getWait())) {
		throw e;
	}	
}

template<class Type>
bool ResourcePool<Type>::free(typename Type::resource_t& plain_res) 
{
	sync_t lk(_condvar);
	typename usedPool_t::iterator i 
		= _usedPool.find(plain_res);
	if (i != _usedPool.end()) {
		resource_ptr res = i->second;
		_usedPool.erase(i);	
		if (res->postcheck()) {
			_freePool.push_back(res);
		} else {
			res->fail_close();
		}
		_condvar.signal(); // notify anyways, to also make the d-tor be awakend
		return true;
	} 
	return false;
}

template<class Type>
bool ResourcePool<Type>::fail(typename Type::resource_t& plain_res)
{
	event(ResourcePoolEventObserver::Event.FAIL_RESOURCE, plain_res.getResourceID()); 
	sync_t lk(_condvar);
	_stats.failCall_();

	typename usedPool_t::iterator i 
		= _usedPool.find(plainresource_ptr(plain_res));
	if (i != _usedPool.end()) {
		resource_ptr res = i->second;
		_usedPool.erase(i);
		res->fail_close();
		return true;
	}
	return false;
}

template<class Type>
void ResourcePool<Type>::_inc_pool()
{
	int PoolSize = _freePool.size() + _usedPool.size();

	if (PoolSize < _options.getMax()) {
		resource_ptr res(_factory.create_resource());
		_freePool.push_back(res);

		std::ostringstream os; os << PoolSize;
		event(ResourcePoolEventObserver::Event.INCREMENT_POOL, res->getResourceID(), os.str());
	} else {
		event(ResourcePoolEventObserver::Event.INCREMENT_POOL, ResourcePoolEventObserver::NoResourceID, "MAX LIMIT REACHED");
		throw CreateException(
			"No more resource available because the Max value has been reached"
		);
	}
}

template<class Type>
void ResourcePool<Type>::sleepit(int tryno) const
{
	return _options.SleepOnFail(tryno);
}

template<class Type>
void ResourcePool<Type>::execute(fatalmind::Command<Type>& cmd) {
	int trynr = getOptions().getMaxExecTry() - 1;
	bool success = false;
	do { 
		typename Type::resource_t& rec = get();
		try {
			cmd.topLevel(true);
			cmd.execute(rec);
			this->free(rec);
			success=true;
		} catch (const NoFailoverException& e) {
			event(ResourcePoolEventObserver::Event.EXECUTE_NFE, rec.getResourceID(), e.message());
			this->free(rec);
			throw;
		} catch (const Exception& e) {
			event(ResourcePoolEventObserver::Event.EXECUTE_FAIL, rec.getResourceID(), e.message());
			fail(rec);
			if (trynr-- <= 0) {
				throw;
			}
		} catch (...) {
			event(ResourcePoolEventObserver::Event.EXECUTE_FAIL, rec.getResourceID());
			fail(rec);
			if (trynr-- <= 0) {
				throw;
			}
		}
	} while (! success);
}

/* not yet
template<class Type>
std::ostream& operator<<(std::ostream& os, const ResourcePool<Type>& p) {
//	typename ResourcePool<Type>::sync_t lk(p._condvar);
		
	typename ResourcePool<Type>::usedPool_t::const_iterator i = p._usedPool.begin();
	os << "ResourcePool@[" << &p << "][USED:[";
	while (i != p._usedPool.end()) {
		os << *(i->second);
		++i;
		if (i != p._usedPool.end()) {
			os << ", ";
		}
	}
	os << "], FREE:[";
	typename ResourcePool<Type>::freePool_t::const_iterator t = p._freePool.begin();
	while (t != p._freePool.end()) {
		os << **t;
		++t;
		if (t != p._freePool.end()) {
			os << ", ";
		}
	}
	return os << "]]";
}
*/


} // namespace fatalmind
