//********************************************************************
//*** ResourcePool/ResourcePoolEventObserver.hh
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: ResourcePoolEventObserver.hh,v 1.18 2009-03-14 13:43:43 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#ifndef RESOURCEPOOL_LOGINTERFACE_HH
#define RESOURCEPOOL_LOGINTERFACE_HH

#include <map>
#include <algorithm>
#include <sstream>
#include "util/Clone.hh"
#include "mm/gc_ptr.hh"
#include "ResourcePool/Command.hh"

namespace fatalmind {

class EventReader; // Fwd decl
/**
 * A ResourcePool Event Observer is a class which get's notified about
 * certain events which happen in the pool. 
 *
 * A classical implementation of an Event Observer is to implement a 
 * logging.
 * On construction of a ResourcePool you can pass a custom derivation
 * of the Event Observer and implement whatever functionality you need.
 *
 * This interface was not designed to interfere with the pool itself,
 * to control the error behaviour of the Pool use the ResourcePoolOptions.
 * The default Observer impelmentation loggs to stderr if enabled.
 * The default is to log NONE events, but the default can be controlled by
 * the FATALMIND_DEFAULT_LOGEVENTS environment variable.
 *
 * e.g. setting \verbatim
FATALMIND_DEFAULT_LOGEVENTS = FAIL_RESOURCE | EXECUTE_FAIL | EXECUTE_NFE \endverbatim
 *
 * would log any of the named failure events.
 *
 * Once you registered your custom implementation you must also
 * call ResourcePool::setLogEvents() and supply the bitmask of events
 * you want to get notifications for.
 * 
 * The instance you pass to the ResourcePool constructor is cloned,
 * the clone is then kept with the ResourcePool. You may destroy
 * the instance you passed to the constructor. 
 * 
 * A very simple, yet useful implementation is the 
 * ResourcePoolStreamLogger which is also used in the \link select.cc select.cc example \endlink.
 * 
 * The following lists the Events available for all ResourcePool's, irregadless of their
 * ResourceType:
 * \li \c INCREMENT_POOL - Indicates that the pool is beeing incremented.
 *      That means that new Resources are constructed and put into the pool of available resources.
 * \li \c DOWNSIZE_POOL - Indicates that the pool is being shruk.
 *      All currently unused resources will be closed.
 * \li \c FAIL_RESOURCE - Indicates that a resource will be fail-closed.
 *      Somebody has returned a resource by passing it to ResourcePool::fail instead of ResourcePool::free.
 * \li \c GET_RESOURCE_FAIL - Indicates that a call to ResourcePool::get failed (did not find a resource).
 * \li \c EXECUTE_NFE - Indicates that a Command execution failed with a NoFailoverException.
 * \li \c EXECUTE_FAIL - Indicates that a Command execution failed with an Exception which is not a NoFailoverException.
 *
 * The following is not really an event, it's used to swtich the timestamp logged to sub-second resolution for logged events:
 * \li \c TIME_HIRES - loggs timestamp with microsecond resolution, but has a little bit more overhead
 *
 * Besides these fundamental events, the following shortcuts are defined:
 * \li \c ALL - All Events (maximum logging), without TIME_HIRES
 * \li \c NONE - No logging.
 * \li \c FAIL - equals \verbatim FAIL_RESOURCE | GET_RESOURCE_FAIL  | EXECUTE_NFE | EXECUTE_FAIL \endverbatim
 *
 * Each ResourceType may define more events, plese refer to the linked topics:
 * \li fatalmind::SQL::SQLFactoryInterface for SQL
 */
class ResourcePoolEventObserver: public Clone {
	public:
		typedef unsigned int event_t;
		static const std::string NoResourceID;

		ResourcePoolEventObserver();

		/**
		 * The callback to notify an observer about an event.
		 *
		 * This method is called by ResourcePool upon the defined events 
		 * (if they are enabled, see ResourcePool::setLogEvents).
		 * The implementation can do whatever it likes, but must take into consideration
		 * that it is synchronously called and might therefore block ResourcePool's processing.
		 *
		 * The call to this method is not protected by any means to avoid reentrant access.
		 * Multi-Thread applications MUST take there own measures to prevent uncoordinated
		 * data manipulation.
		 */
		virtual void event(event_t eventID,  const std::string& resourceID = NoResourceID, const std::string& message ="");

		bool isEvent(event_t eventID) {
			return ((logevents & eventID) != 0);
		}

		/**
		 * Symbol names for the defined events.
		 */
		static struct Events {
			static const event_t INCREMENT_POOL;
			static const event_t DOWNSIZE_POOL;
			static const event_t FAIL_RESOURCE;
			static const event_t GET_RESOURCE_FAIL;
			static const event_t EXECUTE_NFE;
			static const event_t EXECUTE_FAIL;

			static const event_t TIME_HIRES;

			static const event_t FAIL;
			static const event_t ALL;
			static const event_t NONE;
		} Event;

		/**
		 * Translates the numeric EventID to a string representation.
		 */
		// returns references to the members in the map, they become invalid
		// after destruction.
		const std::string& getEventName(const event_t event) const;

		/**
		 * Translates the event name to the EventID. This method is
		 * inteded for seldom usage (since the implementation is ruther
		 * expansive). Use this method to to build the EventID mask which
		 * is set with ResourcePool::setLogEvents.
		 *
		 * @returns The Mapped EventID, or Event.NONE if it could not be mapped.
		 */
		const event_t getEvent(const std::string& value) const;

		/**
		 * Registers a new named event. 
		 * Intended to be used by library authors.
		 * @returns The newly (automatically assigned) EventID 
		 *          or an existing EventID (if the name was already used)
		 *          or Event.NONE if it was not possible to map that event.
		 */
		const event_t registerEvent(const std::string& value) ;

		/**
		 * Translates an EventID Mask expression into it's binary representation.
		 *
		 * <tt>pool.setLogEvents(logger.getEventMask("INCREMENT_POOL | DOWNSIZE_POOL")); </tt>
	 	 * is equivalent (but much slower) then
		 * <tt>pool.setLogEvents(logger.Event.INCREMENT_POOL | logger.Event.DOWNSIZE_POOL)); </tt>
		 */
		const event_t getEventMask(const std::string& value) const;

		inline event_t setLogEvents(event_t events) {
			event_t rv = logevents;
			logevents = events;
			return rv;
		}
		/**
		 * To be used after initialization, will fetch the default from the env.
		 * e.g. after you register new events, to refetch it from the env
		 */
		void setDefaultEventMask();

		gc_ptr<EventReader> getReader(const event_t event, const std::string& r);
		
	protected:

		event_t logevents;
		virtual Clone* DoClone() const;

	private:
		typedef std::map<event_t, std::string> namemap_t;
		namemap_t names;
		int nextEventBit;

		static const std::string INVALID;
};

class EventReader: public usg_cnt_t<DefaultThreadedModel>, public std::ostringstream {
		typedef std::ostringstream super;
	private:
		// todo, could be destructed, would prefere gc_ptr/weak_ptr
		EventReader(ResourcePoolEventObserver& observer, ResourcePoolEventObserver::event_t e, std::string resource);
	friend  class ResourcePoolEventObserver;
	public:
		//EventReader(const EventReader& rho);
		// the d-tor actually loggs
		~EventReader();

/*
		template<typename T>
		inline
		EventReader& operator<< (const std::string& val) {
			if (observer.isEvent(event)) {
				*((super*)this)<<val;
			}
			return *this;
		}
*/

		template<typename T>
		inline
		EventReader& operator<< (const T& val) {
			if (observer.isEvent(event)) {
				super& s = *((super*)this);
				s <<val;
			}
			return *this;
		}

/*
		template<class T>
		EventReader& operator<< (const fatalmind::Command<T>& rho) {
		        if (observer.isEvent(event)) {
		                rho.outputoperator(*this);
		        }
		        return *this;
		}
*/
	private:
		ResourcePoolEventObserver& observer;
		ResourcePoolEventObserver::event_t event;
		std::string resource;
//		std::ostringstream os;
};

} // namespace fatalmind
#endif
