//********************************************************************
//*** ResourcePool/SQL/Factory.cc
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: Factory.cc,v 1.20 2009-05-02 18:47:19 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#include "config.h"
#include "util/Clone.hh"
#include "ResourcePool/CreateException.hh"
#include "ResourcePool/SQL/Factory.hh"
#include "ResourcePool/SQL/Resource.hh"

#ifdef RESOURCEPOOL_HAVE_ORACLE
#include "ResourcePool/oracle/ResourcePool.hh"
#include "ResourcePool/oracle/Resource.hh"
#endif

#ifdef RESOURCEPOOL_HAVE_MYSQL
#include "ResourcePool/mysql/ResourcePool.hh"
#endif

#include <string>
#include <stdexcept>
#include <cassert>

//using namespace std;

namespace fatalmind {
namespace SQL {

/*******
 * constants
 *******/

const string type_oracle_prefix   = "oci";
const string type_oracle9_prefix  = "oci9";
const string type_oracle10_prefix = "oci10";

const string type_mysql_prefix    = "mysql";
const string type_mysql3_prefix   = "mysql3";
const string type_mysql4_prefix   = "mysql4";
const string type_mysql5_prefix   = "mysql5";

/*******
 * Impl decl
 *******/

template<class TM>
class SQLFactory_Impl: public Clone {
	public:
		SQLFactory_Impl(string a_logon)
		: resourcebasedlogon_(a_logon)
		{
		};

		virtual ~SQLFactory_Impl() {
		};

		virtual DummySQLResource<TM>* create_resource() const = 0;

		virtual gc_ptr<ResourcePoolEventObserver> getEventObserver() const = 0;
	private:
		string resourcebasedlogon_;
};


template<class TM, class PT>
class SQLFactory_SpecializedImpl
: public SQLFactory_Impl<TM> {
	public:
		SQLFactory_SpecializedImpl(string a_logon, ResourcePoolEventObserver& e)
		: SQLFactory_Impl<TM>(a_logon)
		, factory_(a_logon, e) 
		{
		};

		virtual ~SQLFactory_SpecializedImpl() {
		};
	
		virtual DummySQLResource<TM>* create_resource() const {
			return new SQLResourceSpecialized<TM, typename PT::resource_t >(
				factory_.create_resource()
			);
		};

		gc_ptr<ResourcePoolEventObserver> getEventObserver() const {
			return factory_.getEventObserver();
		}

	protected:
		virtual Clone* DoClone() const {
			return new SQLFactory_SpecializedImpl<TM,PT>(*this);
		}	
		

	private:
		typename PT::factory_t factory_;
};

/*******
 * Main implementation
 *******/

template<class TM /* a Threading Model */
> 
SQLFactory<TM>::SQLFactory(const string& logon, ResourcePoolEventObserver& e)
: super(e)
, _realFactory()
, resourceTypeCode(UNDEF)
{
	
	string::size_type pos = logon.find(":");
	if (pos == string::npos) {
		string err = "The specified logon string '" 
			+ logon + "' is not valid (no colon found)";
		throw CreateException(err);
	}


	string prefix = logon.substr(0, pos);

	if (false) {
#ifdef RESOURCEPOOL_HAVE_ORACLE
	} else if (
		(prefix == type_oracle_prefix) ||
		(prefix == type_oracle9_prefix) ||
		(prefix == type_oracle10_prefix)
	) {
		resourceTypeCode = OCI;
#endif
#ifdef RESOURCEPOOL_HAVE_MYSQL
	} else if (
		(prefix == type_mysql_prefix)  ||
		(prefix == type_mysql4_prefix) ||
		(prefix == type_mysql5_prefix)
	) {
		resourceTypeCode = MYSQLTX;
	} else if (prefix == type_mysql3_prefix) {
		resourceTypeCode = MYSQL;
#endif
	} else {
		string err = "The specified SQLResource type '" 
			+ prefix 
			+ "' was not compiled in. Available SQLResourcetypes are: " 
// TODO: colon handing ;)
#ifdef RESOURCEPOOL_HAVE_ORACLE
			+ type_oracle_prefix 
#endif
#ifdef RESOURCEPOOL_HAVE_MYSQL
			+ ", " + type_mysql_prefix 
#endif
		;
		throw CreateException(err);
	}
	
	try {
		string resourcebasedlogon = logon.substr(pos+1);
		gc_ptr<SQLFactory_Impl<TM> > impl(
#ifdef RESOURCEPOOL_HAVE_MYSQL
			(resourceTypeCode == MYSQL) || (resourceTypeCode == MYSQLTX) ?
			static_cast<SQLFactory_Impl<TM>* > (
				new SQLFactory_SpecializedImpl<TM, mysql::ResourceType>(resourcebasedlogon, e)
			)
			: 
#endif
#ifdef RESOURCEPOOL_HAVE_ORACLE
			resourceTypeCode == OCI ?
			static_cast<SQLFactory_Impl<TM>* > (
				new SQLFactory_SpecializedImpl<TM, oracle::ResourceType<TM, oracle::SelectListMetaDataCache> >(resourcebasedlogon, e)
			)
			:   
#endif
			throw CreateException("Assertation failure")
		);
		_realFactory = impl;
	} catch (const std::out_of_range& e) {
		string err = "The specified logon '" + logon
			+ " is invalid (missing resource based logon after colon)";
		throw CreateException(err);
	}	
}
template<class TM /* a Threading Model */
> 
SQLFactory<TM>::SQLFactory(const SQLFactory<TM>& rho)
: super(rho)
, _realFactory(rho._realFactory->template clone<SQLFactory_Impl<TM> >())
, resourceTypeCode(rho.resourceTypeCode)
{
}

template<class TM /* a Threading Model */
>
SQLFactory<TM>::~SQLFactory() throw() 
{
}

template<class TM /* a Threading Model */
>
Clone* SQLFactory<TM>::DoClone() const 
{
	return new SQLFactory<TM>(*this);
}

template<class TM /* a Threading Model */
>
DummySQLResource<TM>* SQLFactory<TM>::create_resource() const
{
	return _realFactory->create_resource();
}

template<class TM /* a Threading Model */
>
typename SQLFactory<TM>::databaseType_t SQLFactory<TM>::getDatabaseType() const
{
	return resourceTypeCode;
}

} // namespace SQL
} // namespace fatalmind
