//********************************************************************
//*** ResourcePool/SQL/SQLFactoryInterface.hh
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: SQLFactoryInterface.hh,v 1.14 2009-05-11 14:54:36 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#ifndef RESOURCEPOOL_SQL_SQLFACTORYINTERFACE_HH
#define RESOURCEPOOL_SQL_SQLFACTORYINTERFACE_HH

#include "util/Clone.hh"
#include "mm/gc_ptr.hh"
#include "ResourcePool/ResourcePoolEventObserver.hh"

namespace fatalmind {
namespace SQL {

/** A common Factory interface for all SQL ResourceTypes.
 * 
 * This class defines the following events common for all types of
 * SQL pools:
 * \li \c PREPARE a perperation step before the SQL can be executed. Not all database types require this step.
 * \li \c EXECUTE the actual SQL execution
 * \li \c FETCH when data is fetched from an select SQL
 * \li \c COMMIT when a transaction is ended with a commit
 * \li \c ROLLBACK when a transaction is ended with a rollback
 *
 * Please refer to the driver dependend documentation how to interpret the 4th and following event arguments:
 * \li \c oracleFactory for oracle
 * \li \c mysqlFactory for mysql
 */

class SQLFactoryInterface: public Clone {
	public:
		SQLFactoryInterface(ResourcePoolEventObserver& e) 
		:observer(e.clone<ResourcePoolEventObserver>())
		{
			Event.PREPARE = observer->registerEvent("PREPARE");
			Event.EXECUTE = observer->registerEvent("EXECUTE");
			Event.EXECUTE_FAIL = observer->registerEvent("EXECUTE_FAIL");
			Event.FETCH   = observer->registerEvent("FETCH");
			Event.COMMIT  = observer->registerEvent("COMMIT");
			Event.ROLLBACK= observer->registerEvent("ROLLBACK");

			getEventObserver()->setDefaultEventMask();
		};

		gc_ptr<ResourcePoolEventObserver> getEventObserver() const {
			return observer;
		}

		// TODO: would be nice to have it private/protected?
		enum databaseType_t { UNDEF, OCI, MYSQL, MYSQLTX };
	private:
		struct Events {
			ResourcePoolEventObserver::event_t PREPARE;
			ResourcePoolEventObserver::event_t EXECUTE;
			ResourcePoolEventObserver::event_t EXECUTE_FAIL;
			ResourcePoolEventObserver::event_t FETCH;
			ResourcePoolEventObserver::event_t COMMIT;
			ResourcePoolEventObserver::event_t ROLLBACK;
		} Event;

		const gc_ptr<ResourcePoolEventObserver> observer;
};

} // namespace SQL
} // namespace fatalmind

#endif
