//********************************************************************
//*** ResourcePool/SQL/SQLStatement.hh
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: SQLStatement.hh,v 1.15 2009-05-04 13:20:43 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#ifndef INCLUDE_SQLSTATEMENT_HH
#define INCLUDE_SQLSTATEMENT_HH

#include <string>
//#include <vector>
#include "mm/gc_ptr.hh"

#include "Exceptions/Exception.hh"


namespace fatalmind {
namespace SQL {

class SQLStatementImpl; // Fwd

/**
 * A representation of a SQL string.
 *
 * The SQLStatement class has to be used whenever a SQL statement (string) 
 * has to be passed to a command.
 *
 * Implicit conversion from const char* make expressions like this valid:
 *    \code SQLSelect sel("select * from table"); \endcode
 * In fact, a SQLStatement temporary instance is created and passed to SQLSelect's constructor.
 * 
 * SQLStatement is driver independent, the same instance can be passed to a 
 * oracle or mysql command. The class implements some additional features
 * needed by more then one driver to avoid code duplication. In fact it's current
 * main task is to parse the SQL for placeholders (?) and make meta data available
 * to the drivers. Please note the following additional meta-data which can be 
 * provided for each placeholder:
 * \code
 * ?{<name>}
 * ?{:<type>}
 * ?{<name>:<type>}
 * \endcode
 *
 * That means, a question mark can be followed by curly brackets including a
 * name for the placeholder ([A-Za-z0-9_]) and/or type information (CHAR, INT, FLOAT).
 *
 * Please not that the actual parsing of the SQL string is done upon construction.
 * It's therefore a good practice to have static SQLStatement instances for the 
 * frequently executed statements to avoid parsing overhead.
 * 
 * This information may or may not be used by the actual command, please refer the 
 * the documentation of the command in question.
 * 
 * As of v0.99.27 the information is not used by any command. Future plans are to 
 * allow bind-parameter re-ordering via numbered placeholders (e.g. ?{3}) and
 * named binds. As of v0.99.27, it's not clear if the list of types will remain
 * as it is now, it might also change.
 * 
 * example statements:
 * \code
 * select * from a where a = ? and b = ?
 * select * from a where a = ?{a} and b = ?{b}
 * select * from a where a = ?{a:CHAR} and b = ?{b:INT}
 * \endcode 
 * The first is the classical standard for with a plain question mark as placeholder.
 * The second adds names to each placeholder while the third also supplies type 
 * information to each placeholder.
 *
 * \since 0.99.27
 */
class SQLStatement {
	public:
		/** Holds all available information about a placeholder (?) within an SQL string.
		 */
		struct Parameter {
			// enum? had my troubles with them, don't like them ;(
			static const int UNKNOWN;
			static const int CHAR;
			static const int INT;
			static const int FLOAT;
		};

		/**
		 * Default constructor, same as SQLStatement("").
		 */
		SQLStatement();

		/**
		 * Constructs a new instance from a c++ string.
		 *
		 * \throw SQLException
		 * in case parsing isn't successful. e.g.: 
		 * \code
		 * SQLStatement(std::string("select * from x where a = ?{:INVALID}")); // will throw SQLException
		 * \endcode
		 * The problem is that <tt> ?{:INVALID} </tt> does not name a defined type (\c INVALID instead of \c INT, \c CHAR or \c FLOAT).
		 * Type names are always uppercase, e.g. <tt>?{:int}</tt> would also cause an exception.
		 *
		 */
		SQLStatement(const std::string& sql);

		/**
		 * Constructs a new instance from a c string.
		 * This constructor's main purpose is to not break
		 * pre 0.99.27 code and allow easy in-line SQL's.
		 *
		 * However, you should prefer to construct the SQLStatement
		 * explicitly and keep it to void parsing overhead.
		 * 
		 * Otherwise the same as \see SQLStatement::SQLStatement(const std::string&)
		 */
		SQLStatement(const char* sql);

		SQLStatement(const SQLStatement&);
		~SQLStatement();

		/**
		 * Returns the number of placeholders in the sql string.
		 *
		 * e.g. 
		 * \code
		 * SQLStatement("select * from dual").getParamCount(); // returns 0
		 * SQLStatement("select ? from dual").getParamCount(); // returns 1
		 * SQLStatement("select ? from x where a = ?{a}").getParamCount(); // returns 2
		 * \endcode
		 */
		unsigned int getParamCount() const;

		/**
		 * Returns the name of a placeholder, or the empty string if no name was assigned.
		 * 
		 * e.g. 
		 * \code
		 * SQLStatement("select * from dual").getNameFor(0); // throws std::out_of_range
		 * SQLStatement("select ? from dual").getNameFor(0); // returns an empty string ("")
		 * SQLStatement("select ? from x where a = ?{a}").getNameFor(1); // returns "a"
		 * \endcode
		 *
		 * \throw std::out_of_range if an invalid position is requested.
		 */
		// TODO: name ref?
		std::string getNameFor(const int pos) const;
		
		/**
		 * Returns the type of a placeholder, or the UNKNOWN constant if no type was assigned.
		 * 
		 * e.g. 
		 * \code
		 * SQLStatement("select * from dual").getTypeFor(0); // throws std::out_of_range
		 * SQLStatement("select ? from x where a = ?{a:INT}").getNameFor(0); // returns SQLStatement::Parameter::UNKNOWN
		 * SQLStatement("select ? from x where a = ?{a:INT}").getNameFor(1); // returns SQLStatement::Parameter::INT
		 * \endcode
		 *
		 * \throw std::out_of_range if an invalid position is requested.
		 */
		int getTypeFor(const int pos) const;

		/**
		 * Returns a reference to the original SQL supplied on construction.
		 * 
		 * \return a reference to the member of the SQLStatement instance. 
		 */
		const std::string& getSQL() const;
	
		/**
		 * Returns the n-th fragment of the original SQL when split at each placeholder.
		 *
		 * e.g.
		 * \code
		 * select * from x where a = ? and b = ?{b:INT} order by c
		 * \___________ 0 __________/ \__ 1 __/        \___ 2 ___/
		 * \endcode
		 * Please note that the placeholders themselves are not included.
		 * In case the last character in the SQL is a placeholder, the last 
		 * fragment will be returned as empty string.
		 * \throw std::out_of_range if <tt>pos > (getParamCount() + 1)</tt>
		 *
		 * \return a reference to the member of the SQLStatement instance. 
		 */
		const std::string& getSQLFragment(const int pos) const;

		bool operator==(const SQLStatement& rho);

		bool operator!=(const SQLStatement& rho);

	private:
		// TODO: TM, really important here!
		gc_ptr<SQLStatementImpl>  impl;

		friend class SQLParser;

		SQLStatement(SQLStatementImpl*);

//        no friend? access via getSQL() ;)
//        template<class charT, class traits>
//        friend
//                std::basic_ostream<charT, traits>& operator<<(std::basic_ostream<charT, traits>& s, const SQLStatement& rho);
};

template<class charT, class traits>
std::basic_ostream<charT, traits>& operator<<(std::basic_ostream<charT, traits>& s, const SQLStatement& rho);

} // namespace SQL
} // namespace fatalmind

// required for gc_ptr's Conversion magic to work.
#include "SQLStatementImpl.hh"

#endif
