//********************************************************************
//*** ResourcePool/oracle/StatementHandle.hh
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: OCIHandle.cc,v 1.31 2009-03-18 07:35:10 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#ifndef RESOURCEPOOL_ORACLE_OCIHANDLE_HH
#include "OCIHandle.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_ORACLEEXCEPTIONFACTORY_HH
#include "OracleExceptionFactory.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_RESOURCE_HH
#include "Resource.hh"
#endif

#ifndef OCI_ORACLE
#include <oci.h>
#endif

#include <ostream>

namespace fatalmind {
namespace oracle {

template<typename HTYPE, int HCONST>
OCIHandle<HTYPE, HCONST>::OCIHandle(const OCIEnvHandle& env) 
{
	sword ret;
	void* hdl = &_h;
	if ((ret = OCIHandleAlloc(env.get()
		, (dvoid **)hdl 
		, HCONST
		, 0
		, (dvoid **)0)
	) != OCI_SUCCESS) {
		OracleExceptionFactory::throwit(fatalmind::oracleHandles::LogConstants::OCIHandleAlloc, env, ret, "");
	}
}

template<typename HTYPE, int HCONST>
OCIHandle<HTYPE, HCONST>::OCIHandle(const OCIEnv* env) 
{
	sword ret;
	void* hdl = &_h;
	if ((ret = OCIHandleAlloc(env
		, (dvoid**)hdl
		, HCONST
		, 0
		, (dvoid **)0)
	) != OCI_SUCCESS) {
		OracleExceptionFactory::throwit(fatalmind::oracleHandles::LogConstants::OCIHandleAlloc, env, ret, "");
	}
}

template<typename HTYPE, int HCONST>
OCIHandle<HTYPE, HCONST>::~OCIHandle()
{
	OCIHandleFree((dvoid *)_h, (ub4) HCONST);
}

template class OCIHandle<OCIError,   OCI_HTYPE_ERROR>;
template class OCIHandle<OCIServer,  OCI_HTYPE_SERVER>;
template class OCIHandle<OCISvcCtx,  OCI_HTYPE_SVCCTX>;
template class OCIHandle<OCISession, OCI_HTYPE_SESSION>;
template class OCIHandle<OCIStmt,    OCI_HTYPE_STMT>;

OCIStatementHandle::OCIStatementHandle(const oracleStatement& orasql)
: _SQL(orasql)
, _sth()
, _prepare(0)
, _force_prepare(0)
{
}

bool OCIStatementHandle::prepare(oracleHandles& hand) {
	++_prepare;
	if (_sth.get() == 0) {
		// if it's not prepared yet -> prepare it
		force_prepare(hand);
		return false;
	} else {
		// if it is prepared, check if it is prepared against
		// the right OCIEnv
		OCIEnv* env;
		sword ret;
		if ((ret = OCIAttrGet(_sth->get()
			, OCI_HTYPE_STMT
			, &env
			, 0
			, OCI_ATTR_ENV
			, hand.getError())
		) != OCI_SUCCESS) {
			OracleExceptionFactory::throwit(hand.LogConstant.OCIAttrGet, hand, ret, "");
		}
		if (env != hand.getEnv()) {
			// if not -> prepare new
			force_prepare(hand);
			return false;
		}
	
		// check if it is cacheable, if not
		if (! isCacheable(hand)) {
			// if not -> prepare new
			force_prepare(hand);
			return false;
		}
	}
	return true;
}

void OCIStatementHandle::force_prepare(oracleHandles& hand) {
	++_force_prepare;
	_sth.reset(new Stmt_t(hand.getEnv()));
	sword ret;
	/* prepare statement */
	const std::string& orasql = _SQL.getOraSQL();
	if ((ret = OCIStmtPrepare(_sth->get()
		, hand.getError()
		, (OraText*)orasql.c_str()
		, (ub4) orasql.length()
		, (ub4) OCI_NTV_SYNTAX
		, (ub4) OCI_DEFAULT))
	) {
		OracleExceptionFactory::throwit(hand.LogConstant.OCIStmtPrepare, hand, ret, "");
	}
}

std::string OCIStatementHandle::getOraSQL() const {
	return _SQL.getOraSQL();
}

bool OCIStatementHandle::isCacheable(oracleHandles& hand) const {
	sword ret;
	ub2 stmnt_type;
	if ((ret = OCIAttrGet(_sth->get()
		, OCI_HTYPE_STMT
		, &stmnt_type
		, 0
		, OCI_ATTR_STMT_TYPE	
		, hand.getError()) // TODO, shared error handle?
	) != OCI_SUCCESS) {
		OracleExceptionFactory::throwit(hand.LogConstant.OCIAttrGet, hand, ret, "OCI_ATTR_STMT_TYPE");
	}

	if (stmnt_type == OCI_STMT_SELECT ||
	    stmnt_type == OCI_STMT_UPDATE ||
	    stmnt_type == OCI_STMT_INSERT ||
	    stmnt_type == OCI_STMT_DELETE
	) {
		return true;
	}
	return false;
}

unsigned int OCIStatementHandle::getEfficienty() const 
{
	return _prepare;
	// the _prepare is the efficiency, _force_prepare should be always 1
	// _prepare means the more the better
}

std::ostream& operator<<(std::ostream& s, const OCIStatementHandle& rho) 
{
	return s << '\'' << rho._SQL << '\'' 
		<< " [" << rho._force_prepare << '/' << rho._prepare << ']' << std::endl;
}

} // namespace oracle
} // namespace fatalmind
