//********************************************************************
//*** ResourcePool/oracle/OptimizedBatchCommand.hh
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: OptimizedBatchCommand.cc,v 1.11 2009-05-11 14:06:05 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#include <iostream> // TODO
#include "OptimizedBatchCommand.hh"
#include "ResourcePool/NoopCommandWrapper.hh"

namespace fatalmind {
namespace oracle {

template<class CommandType>
OptimizedBatchCommand<CommandType>::OptimizedBatchCommand(const OptimizedBatchCommand& rho) 
: super(rho)
, lastcmd(0) // TODO: we don't copy the pointer!, howerver this might disable the optimization
{
}

template<class CommandType>
OptimizedBatchCommand<CommandType>::OptimizedBatchCommand(const ResourcePool<typename CommandType::pool_type >& p, bool cont) 
: super(cont)
, lastcmd(0)
{
}

template<class CommandType>
OptimizedBatchCommand<CommandType>::OptimizedBatchCommand(bool cont) 
: super(cont)
, lastcmd(0)
{
}

template<class CommandType>
void OptimizedBatchCommand<CommandType>::add(const CommandType& cmd) {
	const optimizeablecommand_t* ocmd = dynamic_cast<const optimizeablecommand_t*>(&cmd);
	if (ocmd != 0) {
		// if it is an optimzieable command, we call the specialized add method
		// this is required if the exact type is not known at runtime
		add(*ocmd);
	} else {
		// just make sure the flag is set properly and delegate
		lastcmd = 0;
		super::add(cmd);
	}
}

template<class CommandType>
void OptimizedBatchCommand<CommandType>::add(const optimizeablecommand_t& cmd) {
	if (super::doesContinue() && lastcmd != 0 && !super::empty()) {
		if (lastcmd->addBatch(cmd)) {
			// we can benefit from batch execution
			// the actual command is already added to "last" we just need to make sure we push the NoopWrapped command
			NoopCommandWrapper<CommandType> noop(cmd);
			super::add(noop);
			return;
		} else {
			// is not compatible
		}
	} else {
		// either the last command is not of the same type
		// or continuation is disabled.
		// -> in eigther case we can't optimize	
	}
	super::add(cmd);
	// because the vector holds a copy
	typename super::member_t m = super::getLast(); 
	lastcmd = dynamic_cast<optimizeablecommand_t*>(&*m);
}

template<class CommandType>
void OptimizedBatchCommand<CommandType>::execute(typename CommandType::pool_type::resource_t& p) {
	if (super::wasStarted()) {
		typename super::const_iterator lastexecuted = super::getExecPosition();
		const optimizeablecommand_t* ocmd = dynamic_cast<const optimizeablecommand_t*>(lastexecuted->get());
		if (ocmd != 0) {
			// throw previous exception (if any)
			ocmd->throwBatchException();
		}
	}
	try {
		super::execute(p);
	} catch (...) {
		typename super::const_iterator lastexecuted = super::getExecPosition();
		const optimizeablecommand_t* ocmd = dynamic_cast<const optimizeablecommand_t*>(lastexecuted->get());
		if (ocmd != 0) {
			// the statement which has thrown an exception was an Array-executed statement
			// we fetch all error offsets, "normalize" them and add them to our onw list

			// we have to remove the first error offset entry, since it might be wrong
			// the base adds the offset of the batch command, but doesn't know which it actually was
			// however, we need this informaton anyways since it is the offset we need to add

			int lastexecIdx = super::getError();
			int off = super::NO_FURTHER_ERROR; 
			int added = 0;
			while ((off = ocmd->getBatchExceptionOffset())!= super::NO_FURTHER_ERROR ) {
				super::errors.push_back(lastexecIdx + off);
				++added;
			}
			if (!added) {
				// if there was nothing added, we need to re-add the poped one
				// e.g. can happen if a SQLExecute failes which consists of only
				// one singe statement (not batched!)
				// in that case getBatchExceptionOffset() will not work on that stateemet
				super::errors.push_back(lastexecIdx);
			}
		}
		throw;
	}
}

template<class CommandType>
void OptimizedBatchCommand<CommandType>::outputoperator(std::ostream& s) const {
	s << "Optimized";
	super::outputoperator(s);
}

} // namespace oracle
} // namespace fatalmind
