//********************************************************************
//*** ResourcePool/oracle/SQLExecute.cc
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: SQLExecute.cc,v 1.56 2009-05-11 14:54:36 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#include "string.h"
#include "oci.h"
#include <algorithm>

#ifndef RESOURCEPOOL_ORACLE_SQLEXECUTE_HH
#include "SQLExecute.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_RESOURCE_HH
#include "Resource.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_OCIHANDLE_HH
#include "OCIHandle.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_ORACLETYPE_HH
#include "oracleType.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_ORACLEEXCEPTIONFACTORY_HH
#include "OracleExceptionFactory.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_COMMANDINTERFACE_CC
#include "CommandInterface.cc"
#endif

#ifndef RESOURCEPOOL_ORACLE_RESOURCEPOOL_HH
#include "ResourcePool.hh"
#endif

#include <sstream>

#include "ResourcePool/BatchCommand.hh"

namespace fatalmind {
namespace oracle {

template<class TM, template<class> class SLMDC>
const std::string SQLExecute<TM, SLMDC>::_SQLExecuteName("oracle::SQLExecute");


/*
 * The BatchInfo class holds all data required once
 * addBatch is called.
 * It's member of SQLExecute via gc_ptr, which is
 * initialized lazy, to save the overhead
 * in case it's not a batch.
 */

class BatchInfo {
        public:
		BatchInfo(int colcnt) 
		: batchedExceptions()
		, batchedExceptionsFront(0)
		, batchedExceptionOffsets()
		, batchedExceptionOffsetsFront(0)
		, batchParams()
		, batchBuffers()
		, sz_ind()
		{
			// we reserve for batch buffers
			batchBuffers.reserve(colcnt); 
			// and we resize (!) for size and indicator variables
			sz_ind.resize(colcnt);
		}
		// the following store all the excetptions which occured
		// during a batch execution.
		// These structrues are filled during the actual execution
		// of the batch, and the can be accessed afterwards,
		// without the need for an actual resource.
		// two distinct vectors, they are filled at the same time, but read (unqueued) at different times
                mutable std::vector<OracleExceptionFactory> batchedExceptions;
		mutable unsigned int batchedExceptionsFront;
                mutable std::vector<int>                    batchedExceptionOffsets;
		mutable unsigned int batchedExceptionOffsetsFront;

		void reserveBatchExceptions(int rows) {
			batchedExceptions.reserve(rows);
			batchedExceptionOffsets.reserve(rows);
		}
		void addBatchException(OracleExceptionFactory& oef, int off) {
			batchedExceptions.push_back(oef);
			batchedExceptionOffsets.push_back(off);
		}

		void clearExceptions() {
			batchedExceptions.clear();
			batchedExceptionOffsets.clear();
			batchedExceptionsFront = 0;
			batchedExceptionOffsetsFront = 0;
		}

                typedef SQL::Bind<oracleType, oracleTypeImpl>::_argst args_t;
                typedef std::vector<args_t> batchParams_t;
                batchParams_t batchParams;

                std::vector<gc_ptr<RawBuffer> > batchBuffers;

		struct SizeAndIndicator {
			public: SizeAndIndicator(ub2 s, sb2 i) 
			: size(s)
			, indicator(i)
			{
			}
			ub2	size;
			sb2	indicator;

			// to support max_element
			bool operator<(const SizeAndIndicator& rho) const {
				return size < rho.size;
			}
		};
		// first coordinate is the column, the second coordinate is the row
		std::vector<std::vector<struct SizeAndIndicator> > sz_ind;
};

template<class TM, template<class> class SLMDC>
SQLExecute<TM, SLMDC>::SQLExecute(const SQL::SQLStatement& SQL)
: CommandCommon<TM, SLMDC>(SQL)
, batchInfo()
{
}

template<class TM, template<class> class SLMDC>
SQLExecute<TM, SLMDC>::SQLExecute(const ResourcePool<ResourceType<TM> >& p, const SQL::SQLStatement& SQL)
: CommandCommon<TM, SLMDC>(SQL)
, batchInfo()
{
}

template<class TM, template<class> class SLMDC>
SQLExecute<TM, SLMDC>::SQLExecute(const this_t& rho) 
: CommandCommon<TM, SLMDC>(rho)
, batchInfo(rho.batchInfo.get() != 0 ? new BatchInfo(*rho.batchInfo): 0)
{
}

template<class TM, template<class> class SLMDC>
SQLExecute<TM, SLMDC>::~SQLExecute() 
{
}


template<class TM, template<class> class SLMDC>
Clone* SQLExecute<TM, SLMDC>::DoClone() const 
{
	return new SQLExecute<TM, SLMDC>(*this);
};


template<class TM, template<class> class SLMDC>
void SQLExecute<TM, SLMDC>::_execute(oracleResource<TM>& hand, StatementCache& sthcache) 
{
	sword ret;
	_sth = prepare(hand, sthcache, oracleStatement(_SQL));

	if (batchSize() == 1) {
		real_bind(hand, *_sth);
	} else {
		real_bind_batch(hand);

		// reset all stored execption from prior batch executions.
		// just for the case that the batch is executed multiple times
		batchInfo->clearExceptions();
	}


	// if there is no surrounding transaction, we always commit
	bool commit_on_success = (hand.getTxDepth() == 0);
	/* execute statement */
	if ((ret = OCIStmtExecute(hand.getSvcCtx()
		, _sth->get()
		, hand.getError()
		, (ub4) batchSize()
		, (ub4) 0
		, (CONST OCISnapshot *) 0
		, (OCISnapshot *) 0
		, (ub4) (commit_on_success ? OCI_COMMIT_ON_SUCCESS : OCI_DEFAULT) | (batchSize() > 1 ? OCI_BATCH_ERRORS :  OCI_DEFAULT)) 
	) != OCI_SUCCESS)
	{
		if (batchSize() == 1) { 
			_sth.reset(); // not needed anymore, if reused-cashed version is taken anyways
			OracleExceptionFactory oef(hand.LogConstant.OCIStmtExecute, hand.getError(), ret, _SQL);
			gc_ptr<EventReader> er = hand.getReader(hand.Event.EXECUTE_FAIL);
			(*er) << oef.getErrorCode() << hand.LogConstant.SEMICOLON << (*this); 
			if (batchInfo.get() != 0) {
				batchInfo->addBatchException(oef, 0);
			}
			oef.throwit();
		} else {
			handleBatchException(ret, hand);
			throwBatchException();
		}
	} 
	(*hand.getReader(hand.Event.EXECUTE)) <<  hand.LogConstant.SEMICOLON  << (*this);	
	if (commit_on_success) {
		hand.event(hand.Event.COMMIT, hand.LogConstant.IMPLICIT);
	}
	_sth.reset(); // not needed anymore, if reused-cashed version is taken anyways
}

template<class TM, template<class> class SLMDC>
void SQLExecute<TM, SLMDC>::handleBatchException(int ret, oracleResource<TM>& hand) {
	gc_ptr<EventReader> er = hand.getReader(hand.Event.EXECUTE_FAIL);
	int row_offset = 42, num_errors;
	OCIErrorHandle errhp2(hand.getEnv());

	OCIAttrGet (_sth->get(), OCI_HTYPE_STMT, &num_errors, 0,  OCI_ATTR_NUM_DML_ERRORS, errhp2.get());
	(*er) << "BATCH ERROR;" << num_errors << " of " << batchSize() << " failed: ";
	OCIErrorHandle err(hand.getEnv());
	void *e = err.get();
	batchInfo->reserveBatchExceptions(num_errors ? num_errors : batchSize());
	for (int i = 0; i < num_errors; ++i) {
		OCIParamGet(hand.getError(), OCI_HTYPE_ERROR, errhp2.get(), &e, i);

		OCIAttrGet(err.get(), OCI_HTYPE_ERROR, &row_offset, 0, OCI_ATTR_DML_ROW_OFFSET, errhp2.get());

		std::ostringstream oss;
		oss << "OCIStmtExecute(at batch position " << row_offset << " with " << num_errors << " errors)";
		
		OracleExceptionFactory oef(hand.LogConstant.OCIStmtExecute, err.get(), ret,  OCI_HTYPE_ERROR, _SQL);
		batchInfo->addBatchException(oef, row_offset);
		(*er) << "[Pos=" << row_offset << ";" << oss.str() << ":" << oef.getErrorString() << ']';
	}
	if (num_errors == 0) {
		// if not a single command was executed successfully (e.g. because of a syntax error)
		// ret code of StmtExecute is -1 (no successful) but num errors is 0,
		// we need to "emulate" the errors
		for (int i = 0; i < batchSize(); ++i) {
			OracleExceptionFactory oef(hand.LogConstant.OCIStmtExecute, hand.getError(), ret, OCI_HTYPE_ERROR, _SQL);
			batchInfo->addBatchException(oef, i);
			(*er) << "[Pos=" << row_offset << ";" << oef.getErrorString() << ']';
		}
	}
	_sth.reset(); // not needed anymore, if reused-cashed version is taken anyways
}


template<class TM, template<class> class SLMDC>
void SQLExecute<TM, SLMDC>::throwBatchException() const {
	if (batchInfo.get() != 0 && batchInfo->batchedExceptions.size() > batchInfo->batchedExceptionsFront) {
			OracleExceptionFactory oef(batchInfo->batchedExceptions[batchInfo->batchedExceptionsFront++]);
			oef.throwit();
	}
}

template<class TM, template<class> class SLMDC>
int SQLExecute<TM, SLMDC>::getBatchExceptionOffset() const {
	if (batchInfo.get() != 0 && batchInfo->batchedExceptionOffsets.size() > batchInfo->batchedExceptionOffsetsFront) {
			int rv = batchInfo->batchedExceptionOffsets[batchInfo->batchedExceptionOffsetsFront++];
			return rv;
	}
	return BatchCommand<Command<ResourceType<TM> > >::NO_FURTHER_ERROR;
}


template<class TM, template<class> class SLMDC>
void SQLExecute<TM, SLMDC>::real_bind_batch(oracleResource<TM>& hand) {
	sword ret;
	typename super::_argst::const_iterator end = super::_args.end();
        int pos = 0;
	batchInfo->batchBuffers.clear();
        for (typename super::_argst::const_iterator i = super::_args.begin(); i != end; ++i) {
		// obtain maximum size
		int mxsize = max_element(batchInfo->sz_ind[pos].begin(), batchInfo->sz_ind[pos].end())->size;

		gc_ptr<RawBuffer> buf(new RawBuffer(mxsize * batchInfo->batchParams.size()));
		batchInfo->batchBuffers.push_back(buf); // todo, release
		
		// copy the data into the buffer
		typename BatchInfo::batchParams_t::const_iterator tend = batchInfo->batchParams.end();	
		char *buffer = (char*)buf->get();
		for (typename BatchInfo::batchParams_t::const_iterator t = batchInfo->batchParams.begin(); t != tend; ++t) {
			memcpy(buffer, (*t)[pos]->ptr(), (*t)[pos]->size());
			buffer += mxsize;
		}

		OCIBind  *bnd1p = NULL;
		if ((ret = OCIBindByPos(_sth->get()
			, &bnd1p
			, hand.getError()
			, pos+1
			, buf->get()
			, mxsize
			, (*i)->type()
			, &(batchInfo->sz_ind[pos][0].indicator)
			, &(batchInfo->sz_ind[pos][0].size) // alenp
			, (ub2 *)0
			, 0 // maxarr_len (only for PL/SQL)
			, (ub4 *)0
			, OCI_DEFAULT)
		) != OCI_SUCCESS) {
			OracleExceptionFactory::throwit(hand.LogConstant.OCIBindByPos, hand, ret, _SQL);
		}


		// array it ;)
		if ((ret = OCIBindArrayOfStruct(bnd1p
			, hand.getError()
			, mxsize // pvskip
//			, sizeof(sb2) // indskip
//			, sizeof(ub2) // actual length skip parameter
			, sizeof(struct BatchInfo::SizeAndIndicator) // indicator skip
			, sizeof(struct BatchInfo::SizeAndIndicator) // size (alenp) skip
			, 0)
		) != OCI_SUCCESS) {
			OracleExceptionFactory::throwit(hand.LogConstant.OCIBindArrayOfStruct, hand, ret, _SQL);
		}
                ++pos;
        }
}

template<class TM, template<class> class SLMDC>
int SQLExecute<TM, SLMDC>::batchSize() const {
	if (batchInfo.get() == 0) {
		return 1;
	} else {
		return batchInfo->batchParams.size();
	}
}

template<class TM, template<class> class SLMDC>
bool SQLExecute<TM, SLMDC>::push_onto_BatchParamSizes(const typename super::super_bind::_argst& p) {
	if (batchInfo->sz_ind.size() != p.size()) {
		std::ostringstream os;
		os << "Error in batch execution: number of arguments mismatch. Current batch has " << batchInfo->sz_ind.size()  << " while the supplied statement has " << p.size();
		throw SQL::SQLException(os.str(), "07001", super::getSQL(), false);
		// note on usage of getSQL: when reaching this point it is already verified that this SQL equals that SQL
	}

	// we need to iterate twice..
	// first to check consistance of the actual types, without modifying any data in *this.
	//       in case they are not the same, we can still abort batching without 
	//       having corrupted data
	// TODO, alternativley we could "collback" by "truncating" the rows from the arrays
	typename super::_argst::const_iterator end = p.end();
	typename super::_argst::const_iterator t =  super::_args.begin();
	for (typename super::super_bind::_argst::const_iterator i = p.begin(); i != end; ++i) {
		if (((*i).get() != 0) && ((*t).get() != 0) and (*i)->type() != (*t)->type()) {
			// log an warning someting like:
			// std::cout << "Can not batch up similar statements due to different parameter type at position " << distance(static_cast<typename super::super_bind::_argst::const_iterator>(p.begin()), i) << " (" << ((*i)->type()) <<  " vs. " <<  ((*t)->type()) << ")" << std::endl;
			return false;
		}
		++t;
	}
	copySizeIndicator(p);
	return true;
}

template<class TM, template<class> class SLMDC>
void SQLExecute<TM, SLMDC>::copySizeIndicator(const typename super::super_bind::_argst& p) {
	typename super::_argst::const_iterator end = p.end();
        int pos = 0;
	for (typename super::super_bind::_argst::const_iterator i = p.begin(); i != end; ++i) {
		batchInfo->sz_ind[pos].push_back(BatchInfo::SizeAndIndicator((*i)->size(), (*i)->indicator()));
		++pos;	
	}
}

template<class TM, template<class> class SLMDC>
bool SQLExecute<TM, SLMDC>::addBatch(const SQLExecute<TM, SLMDC>& p) {
	if (_sth.get()) {
		// TODO: recheck
		throw SQL::SQLException("building up a batch for a already prepared statement is invalid", "24000", p._SQL, false);
	}
	if (_SQL != p._SQL) {
		//std::cout << " false due to different sql (" << _SQL << ") vs. (" << p._SQL << ")" << std::endl;
		return false;
	}

	putInsideBatch();

	if (push_onto_BatchParamSizes(p._args)) {
		batchInfo->batchParams.push_back(p._args);
		return true;
	} else {
		return false;
	}
}
template<class TM, template<class> class SLMDC>
const std::string& SQLExecute<TM, SLMDC>::getCommandName() const {
	return _SQLExecuteName;
}

template<class TM, template<class> class SLMDC>
void SQLExecute<TM, SLMDC>::bindoutputoperator(std::ostream& str) const {
	if (batchSize() == 1) {
		super::bindoutputoperator(str);
	} else {
		typename BatchInfo::batchParams_t::const_iterator e =  batchInfo->batchParams.end();
		for (typename BatchInfo::batchParams_t::const_iterator i = batchInfo->batchParams.begin(); i != e; ++i) {
			super::super_bind::bindoutputoperator(str, *i);
		}
	}
}

template<class TM, template<class> class SLMDC>
void SQLExecute<TM, SLMDC>::putInsideBatch() {
	if (batchInfo.get() == 0) {
		// if this is the first call to addBatch, 
		// first we initialize the batchInfo structure
		batchInfo.reset(new BatchInfo(super::super_bind::_args.size()));

		// then we copy the already
		// (normally) bound values into the first positon of our new structure
		batchInfo->batchParams.push_back(super::super_bind::_args);
		copySizeIndicator(super::super_bind::_args);
	}
}

template class SQLExecute<DefaultThreadedModel, oracle::SelectListMetaDataCache >;
//template class SQLExecute<DefaultThreadedModel, oracle::SelectListMetaDataNopCache >;

} // namespace oracle
} // namespace fatalmind

